'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var syntaxJsx = _interopDefault(require('@babel/plugin-syntax-jsx'));
var helperModuleImports = require('@babel/helper-module-imports');
var htmlTags = _interopDefault(require('html-tags'));
var svgTags = _interopDefault(require('svg-tags'));

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

const PatchFlags = {
  TEXT: 1,
  CLASS: 1 << 1,
  STYLE: 1 << 2,
  PROPS: 1 << 3,
  FULL_PROPS: 1 << 4,
  HYDRATE_EVENTS: 1 << 5,
  STABLE_FRAGMENT: 1 << 6,
  KEYED_FRAGMENT: 1 << 7,
  UNKEYED_FRAGMENT: 1 << 8,
  NEED_PATCH: 1 << 9,
  DYNAMIC_SLOTS: 1 << 10,
  HOISTED: -1,
  BAIL: -2
}; // dev only flag -> name mapping

const PatchFlagNames = {
  [PatchFlags.TEXT]: 'TEXT',
  [PatchFlags.CLASS]: 'CLASS',
  [PatchFlags.STYLE]: 'STYLE',
  [PatchFlags.PROPS]: 'PROPS',
  [PatchFlags.FULL_PROPS]: 'FULL_PROPS',
  [PatchFlags.HYDRATE_EVENTS]: 'HYDRATE_EVENTS',
  [PatchFlags.STABLE_FRAGMENT]: 'STABLE_FRAGMENT',
  [PatchFlags.KEYED_FRAGMENT]: 'KEYED_FRAGMENT',
  [PatchFlags.UNKEYED_FRAGMENT]: 'UNKEYED_FRAGMENT',
  [PatchFlags.NEED_PATCH]: 'NEED_PATCH',
  [PatchFlags.DYNAMIC_SLOTS]: 'DYNAMIC_SLOTS',
  [PatchFlags.HOISTED]: 'HOISTED',
  [PatchFlags.BAIL]: 'BAIL'
};

const createIdentifier = (t, state, id) => t.memberExpression(state.get('vue'), t.identifier(id));
/**
 * Checks if string is describing a directive
 * @param src string
 */


const isDirective = src => src.startsWith('v-') || src.startsWith('v') && src.length >= 2 && src[1] >= 'A' && src[1] <= 'Z';
/**
 * Check if a JSXOpeningElement is fragment
 * @param {*} t
 * @param {*} path
 * @returns boolean
 */


const isFragment = (t, path) => t.isJSXMemberExpression(path) && path.node.property.name === 'Fragment';
/**
 * Check if a JSXOpeningElement is a component
 *
 * @param t
 * @param path JSXOpeningElement
 * @returns boolean
 */


const checkIsComponent = (t, path) => {
  const namePath = path.get('name');

  if (t.isJSXMemberExpression(namePath)) {
    return !isFragment(t, namePath); // For withCtx
  }

  const tag = namePath.get('name').node;
  return !htmlTags.includes(tag) && !svgTags.includes(tag);
};
/**
 * Transform JSXMemberExpression to MemberExpression
 * @param t
 * @param path JSXMemberExpression
 * @returns MemberExpression
 */


const transformJSXMemberExpression = (t, path) => {
  const objectPath = path.get('object');
  const propertyPath = path.get('property');
  const transformedObject = objectPath.isJSXMemberExpression() ? transformJSXMemberExpression(t, objectPath) : objectPath.isJSXIdentifier() ? t.identifier(objectPath.node.name) : t.nullLiteral();
  const transformedProperty = t.identifier(propertyPath.get('name').node);
  return t.memberExpression(transformedObject, transformedProperty);
};
/**
 * Get tag (first attribute for h) from JSXOpeningElement
 * @param t
 * @param path JSXOpeningElement
 * @returns Identifier | StringLiteral | MemberExpression
 */


const getTag = (t, path) => {
  const namePath = path.get('openingElement').get('name');

  if (namePath.isJSXIdentifier()) {
    const {
      name
    } = namePath.node;

    if (path.scope.hasBinding(name) && !htmlTags.includes(name) && !svgTags.includes(name)) {
      return t.identifier(name);
    }

    return t.stringLiteral(name);
  }

  if (namePath.isJSXMemberExpression()) {
    return transformJSXMemberExpression(t, namePath);
  }

  throw new Error(`getTag: ${namePath.type} is not supported`);
};

const getJSXAttributeName = (t, path) => {
  const nameNode = path.node.name;

  if (t.isJSXIdentifier(nameNode)) {
    return nameNode.name;
  }

  return `${nameNode.namespace.name}:${nameNode.name.name}`;
};
/**
 * Transform JSXText to StringLiteral
 * @param t
 * @param path JSXText
 * @returns StringLiteral
 */


const transformJSXText = (t, path) => {
  const {
    node
  } = path;
  const lines = node.value.split(/\r\n|\n|\r/);
  let lastNonEmptyLine = 0;

  for (let i = 0; i < lines.length; i++) {
    if (lines[i].match(/[^ \t]/)) {
      lastNonEmptyLine = i;
    }
  }

  let str = '';

  for (let i = 0; i < lines.length; i++) {
    const line = lines[i];
    const isFirstLine = i === 0;
    const isLastLine = i === lines.length - 1;
    const isLastNonEmptyLine = i === lastNonEmptyLine; // replace rendered whitespace tabs with spaces

    let trimmedLine = line.replace(/\t/g, ' '); // trim whitespace touching a newline

    if (!isFirstLine) {
      trimmedLine = trimmedLine.replace(/^[ ]+/, '');
    } // trim whitespace touching an endline


    if (!isLastLine) {
      trimmedLine = trimmedLine.replace(/[ ]+$/, '');
    }

    if (trimmedLine) {
      if (!isLastNonEmptyLine) {
        trimmedLine += ' ';
      }

      str += trimmedLine;
    }
  }

  return str !== '' ? t.stringLiteral(str) : null;
};
/**
 * Transform JSXExpressionContainer to Expression
 * @param path JSXExpressionContainer
 * @returns Expression
 */


const transformJSXExpressionContainer = path => path.get('expression').node;
/**
 * Transform JSXSpreadChild
 * @param t
 * @param path JSXSpreadChild
 * @returns SpreadElement
 */


const transformJSXSpreadChild = (t, path) => t.spreadElement(path.get('expression').node);
/**
 * Get JSX element type
 *
 * @param t
 * @param path Path<JSXOpeningElement>
 */


const getType = (t, path) => {
  const typePath = path.get('attributes').find(attributePath => t.isJSXAttribute(attributePath) && t.isJSXIdentifier(attributePath.get('name')) && attributePath.get('name.name').node === 'type' && t.isStringLiteral(attributePath.get('value')));
  return typePath ? typePath.get('value.value').node : '';
};

const resolveDirective = (t, path, state, tag, directiveName) => {
  if (directiveName === 'show') {
    return createIdentifier(t, state, 'vShow');
  }

  if (directiveName === 'model') {
    let modelToUse;
    const type = getType(t, path.parentPath);

    switch (tag.value) {
      case 'select':
        modelToUse = createIdentifier(t, state, 'vModelSelect');
        break;

      case 'textarea':
        modelToUse = createIdentifier(t, state, 'vModelText');
        break;

      default:
        switch (type) {
          case 'checkbox':
            modelToUse = createIdentifier(t, state, 'vModelCheckbox');
            break;

          case 'radio':
            modelToUse = createIdentifier(t, state, 'vModelRadio');
            break;

          default:
            modelToUse = createIdentifier(t, state, 'vModelText');
        }

    }

    return modelToUse;
  }

  return t.callExpression(createIdentifier(t, state, 'resolveDirective'), [t.stringLiteral(directiveName)]);
};
/**
 * Parse directives metadata
 *
 * @param t
 * @param  path JSXAttribute
 * @returns null | Object<{ modifiers: Set<string>, valuePath: Path<Expression>}>
 */


const parseDirectives = (t, {
  name,
  path,
  value,
  state,
  tag,
  isComponent
}) => {
  const modifiers = name.split('_');
  const directiveName = modifiers.shift().replace(/^v/, '').replace(/^-/, '').replace(/^\S/, s => s.toLowerCase());

  if (directiveName === 'model' && !t.isJSXExpressionContainer(path.get('value'))) {
    throw new Error('You have to use JSX Expression inside your v-model');
  }

  const modifiersSet = new Set(modifiers);
  const hasDirective = directiveName !== 'model' || directiveName === 'model' && !isComponent;
  return {
    directiveName,
    modifiers: modifiersSet,
    directive: hasDirective ? [resolveDirective(t, path, state, tag, directiveName), value, modifiersSet.size && t.unaryExpression('void', t.numericLiteral(0), true), modifiersSet.size && t.objectExpression([...modifiersSet].map(modifier => t.objectProperty(t.identifier(modifier), t.booleanLiteral(true))))].filter(Boolean) : undefined
  };
};

const xlinkRE = /^xlink([A-Z])/;
const onRE = /^on[^a-z]/;

const isOn = key => onRE.test(key);

const transformJSXSpreadAttribute = (t, path, mergeArgs) => {
  const argument = path.get('argument').node;
  const {
    properties
  } = argument;

  if (!properties) {
    // argument is an Identifier
    mergeArgs.push(argument);
  } else {
    mergeArgs.push(t.objectExpression(properties));
  }
};

const getJSXAttributeValue = (t, path, state) => {
  const valuePath = path.get('value');

  if (valuePath.isJSXElement()) {
    return transformJSXElement(t, valuePath, state);
  }

  if (valuePath.isStringLiteral()) {
    return valuePath.node;
  }

  if (valuePath.isJSXExpressionContainer()) {
    return transformJSXExpressionContainer(valuePath);
  }

  return null;
};
/**
 *  Check if an attribute value is constant
 * @param t
 * @param path
 * @returns boolean
 */


const isConstant = (t, path) => {
  if (t.isIdentifier(path)) {
    return path.name === 'undefined';
  }

  if (t.isArrayExpression(path)) {
    return path.elements.every(element => isConstant(t, element));
  }

  if (t.isObjectExpression(path)) {
    return path.properties.every(property => isConstant(t, property.value));
  }

  if (t.isLiteral(path)) {
    return true;
  }

  return false;
};

const mergeAsArray = (t, existing, incoming) => {
  if (t.isArrayExpression(existing.value)) {
    existing.value.elements.push(incoming.value);
  } else {
    existing.value = t.arrayExpression([existing.value, incoming.value]);
  }
};

const dedupeProperties = (t, properties = []) => {
  const knownProps = new Map();
  const deduped = [];
  properties.forEach(prop => {
    const {
      key: {
        value: name
      } = {}
    } = prop;
    const existing = knownProps.get(name);

    if (existing) {
      if (name === 'style' || name === 'class' || name.startsWith('on')) {
        mergeAsArray(t, existing, prop);
      }
    } else {
      knownProps.set(name, prop);
      deduped.push(prop);
    }
  });
  return deduped;
};

const buildProps = (t, path, state, hasContainer) => {
  const tag = getTag(t, path);
  const isComponent = checkIsComponent(t, path.get('openingElement'));
  const props = path.get('openingElement').get('attributes');
  const directives = [];
  const dynamicPropNames = new Set();
  let patchFlag = 0;

  if (isFragment(t, path.get('openingElement.name'))) {
    patchFlag |= PatchFlags.STABLE_FRAGMENT;
  } else if (hasContainer) {
    patchFlag |= PatchFlags.BAIL;
  }

  if (props.length === 0) {
    return {
      tag,
      props: t.nullLiteral(),
      directives,
      patchFlag,
      dynamicPropNames
    };
  }

  const properties = []; // patchFlag analysis

  let hasRef = false;
  let hasClassBinding = false;
  let hasStyleBinding = false;
  let hasHydrationEventBinding = false;
  let hasDynamicKeys = false;
  const mergeArgs = [];
  props.forEach(prop => {
    if (prop.isJSXAttribute()) {
      let name = getJSXAttributeName(t, prop);
      const attributeValue = getJSXAttributeValue(t, prop);

      if (!isConstant(t, attributeValue) || name === 'ref') {
        if (!isComponent && isOn(name) // omit the flag for click handlers becaues hydration gives click
        // dedicated fast path.
        && name.toLowerCase() !== 'onclick' // omit v-model handlers
        && name !== 'onUpdate:modelValue') {
          hasHydrationEventBinding = true;
        }

        if (name === 'ref') {
          hasRef = true;
        } else if (name === 'class' && !isComponent) {
          hasClassBinding = true;
        } else if (name === 'style' && !isComponent) {
          hasStyleBinding = true;
        } else if (name !== 'key' && !isDirective(name) && name !== 'on') {
          dynamicPropNames.add(name);
        }
      }

      if (state.opts.transformOn && (name === 'on' || name === 'nativeOn')) {
        if (!state.get('transformOn')) {
          state.set('transformOn', helperModuleImports.addDefault(path, '@ant-design-vue/babel-helper-vue-transform-on', {
            nameHint: '_transformOn'
          }));
        }

        mergeArgs.push(t.callExpression(state.get('transformOn'), [attributeValue || t.booleanLiteral(true)]));
        return;
      }

      if (isDirective(name)) {
        const {
          directive,
          modifiers,
          directiveName
        } = parseDirectives(t, {
          tag,
          isComponent,
          name,
          path: prop,
          state,
          value: attributeValue
        });

        if (directive) {
          directives.push(t.arrayExpression(directive));
        } else {
          // must be v-model and is a component
          properties.push(t.objectProperty(t.stringLiteral('modelValue'), attributeValue));
          dynamicPropNames.add('modelValue');

          if (modifiers.size) {
            properties.push(t.objectProperty(t.stringLiteral('modelModifiers'), t.objectExpression([...modifiers].map(modifier => t.objectProperty(t.stringLiteral(modifier), t.booleanLiteral(true))))));
          }
        }

        if (directiveName === 'model') {
          properties.push(t.objectProperty(t.stringLiteral('onUpdate:modelValue'), t.arrowFunctionExpression([t.identifier('$event')], t.assignmentExpression('=', attributeValue, t.identifier('$event')))));
          dynamicPropNames.add('onUpdate:modelValue');
        }

        return;
      }

      if (name.match(xlinkRE)) {
        name = name.replace(xlinkRE, (_, firstCharacter) => `xlink:${firstCharacter.toLowerCase()}`);
      }

      properties.push(t.objectProperty(t.stringLiteral(name), attributeValue || t.booleanLiteral(true)));
    } else {
      hasDynamicKeys = true;
      transformJSXSpreadAttribute(t, prop, mergeArgs);
    }
  }); // patchFlag analysis

  if (hasDynamicKeys) {
    patchFlag |= PatchFlags.FULL_PROPS;
  } else {
    if (hasClassBinding) {
      patchFlag |= PatchFlags.CLASS;
    }

    if (hasStyleBinding) {
      patchFlag |= PatchFlags.STYLE;
    }

    if (dynamicPropNames.size) {
      patchFlag |= PatchFlags.PROPS;
    }

    if (hasHydrationEventBinding) {
      patchFlag |= PatchFlags.HYDRATE_EVENTS;
    }
  }

  if ((patchFlag === 0 || patchFlag === PatchFlags.HYDRATE_EVENTS) && hasRef) {
    patchFlag |= PatchFlags.NEED_PATCH;
  }

  let propsExpression;

  if (mergeArgs.length) {
    if (properties.length) {
      mergeArgs.push(...dedupeProperties(t, properties));
    }

    if (mergeArgs.length > 1) {
      const exps = [];
      const objectProperties = [];
      mergeArgs.forEach(arg => {
        if (t.isIdentifier(arg) || t.isExpression(arg)) {
          exps.push(arg);
        } else {
          objectProperties.push(arg);
        }
      });
      propsExpression = t.callExpression(createIdentifier(t, state, 'mergeProps'), [...exps, objectProperties.length && t.objectExpression(objectProperties)].filter(Boolean));
    } else {
      // single no need for a mergeProps call
      // eslint-disable-next-line prefer-destructuring
      propsExpression = mergeArgs[0];
    }
  } else if (properties.length) {
    propsExpression = t.objectExpression(dedupeProperties(t, properties));
  }

  return {
    tag,
    props: propsExpression,
    directives,
    patchFlag,
    dynamicPropNames
  };
};
/**
 * Get children from Array of JSX children
 * @param t
 * @param paths Array<JSXText | JSXExpressionContainer | JSXSpreadChild | JSXElement>
 * @returns Array<Expression | SpreadElement>
 */


const getChildren = (t, paths, state) => {
  let hasContainer = false;
  return {
    children: paths.map(path => {
      if (path.isJSXText()) {
        return transformJSXText(t, path);
      }

      if (path.isJSXExpressionContainer()) {
        hasContainer = true;
        return transformJSXExpressionContainer(path);
      }

      if (path.isJSXSpreadChild()) {
        return transformJSXSpreadChild(t, path);
      }

      if (path.isCallExpression()) {
        return path.node;
      }

      if (path.isJSXElement()) {
        return transformJSXElement(t, path, state);
      }

      throw new Error(`getChildren: ${path.type} is not supported`);
    }).filter(value => value !== undefined && value !== null && !t.isJSXEmptyExpression(value)),
    hasContainer
  };
};

const transformJSXElement = (t, path, state) => {
  const {
    children,
    hasContainer
  } = getChildren(t, path.get('children'), state);
  const {
    tag,
    props,
    directives,
    patchFlag,
    dynamicPropNames
  } = buildProps(t, path, state, hasContainer);
  const flagNames = Object.keys(PatchFlagNames).map(Number).filter(n => n > 0 && patchFlag & n).map(n => PatchFlagNames[n]).join(', ');
  const isComponent = checkIsComponent(t, path.get('openingElement'));
  const child = children.length === 1 && t.isStringLiteral(children[0]) ? children[0] : t.arrayExpression(children);

  if (state.opts.compatibleProps && !state.get('compatibleProps')) {
    state.set('compatibleProps', helperModuleImports.addDefault(path, '@ant-design-vue/babel-helper-vue-compatible-props', {
      nameHint: '_compatibleProps'
    }));
  }

  const createVNode = t.callExpression(createIdentifier(t, state, 'createVNode'), [tag, state.opts.compatibleProps ? t.callExpression(state.get('compatibleProps'), [props]) : props, children[0] ? isComponent ? t.objectExpression([t.objectProperty(t.identifier('default'), t.callExpression(createIdentifier(t, state, 'withCtx'), [t.arrowFunctionExpression([], t.isStringLiteral(child) ? t.callExpression(createIdentifier(t, state, 'createTextVNode'), [child]) : child)])), t.objectProperty(t.identifier('_'), t.numericLiteral(1))]) : child : t.nullLiteral(), patchFlag && t.addComment(t.numericLiteral(patchFlag), 'trailing', ` ${flagNames} `, false), dynamicPropNames.size && t.arrayExpression([...dynamicPropNames.keys()].map(name => t.stringLiteral(name)))].filter(Boolean));

  if (!directives.length) {
    return createVNode;
  }

  return t.callExpression(createIdentifier(t, state, 'withDirectives'), [createVNode, t.arrayExpression(directives)]);
};

var tranformVueJSX = (t => ({
  JSXElement: {
    exit(path, state) {
      if (!state.get('vue')) {
        state.set('vue', helperModuleImports.addNamespace(path, 'vue'));
      }

      path.replaceWith(transformJSXElement(t, path, state));
    }

  }
}));

const transformFragment = (t, path, Fragment) => {
  const children = path.get('children') || [];
  return t.jsxElement(t.jsxOpeningElement(Fragment, []), t.jsxClosingElement(Fragment), children.map(({
    node
  }) => node), false);
};

var sugarFragment = (t => ({
  JSXFragment: {
    enter(path, state) {
      if (!state.get('vue')) {
        state.set('vue', helperModuleImports.addNamespace(path, 'vue'));
      }

      path.replaceWith(transformFragment(t, path, t.jsxMemberExpression(t.jsxIdentifier(state.get('vue').name), t.jsxIdentifier('Fragment'))));
    }

  }
}));

var index = (({
  types: t
}) => ({
  name: 'babel-plugin-jsx',
  inherits: syntaxJsx,
  visitor: _objectSpread2(_objectSpread2({}, tranformVueJSX(t)), sugarFragment(t))
}));

module.exports = index;
